module Evaluation.FunctionIO
  ( evalFunctionIO
  ) where

import Prelude hiding (String, FilePath)

import Types.Values
import Types (FunctionIO(..), Value(..), toValue)

import qualified Data.Text.Lazy as TL
import qualified Data.Text.IO as T
import qualified Text.Pandoc as P
import qualified Text.Blaze.Html.Renderer.Text as B
import qualified Codec.Picture as CP
import qualified Codec.Picture.STBIR as CPS
import System.Directory (listDirectory, doesDirectoryExist, createDirectory, copyFile)

evalFunctionIO :: FunctionIO -> Value -> IO Value
evalFunctionIO f x = case (f, x) of
  (ListDirectory, String (StringWrapper s)) ->
    (List . map (toValue . StringWrapper)) <$> listDirectory s

  (IsDirectory, String (StringWrapper s)) ->
    Bool <$> doesDirectoryExist s

  (ReadTextFile, String (StringWrapper s)) ->
    Text <$> T.readFile s

  (ConvertImage, Tuple (Tuple (String (StringWrapper source), String (StringWrapper target)),
                        ImageConversionSettings (ResizeToWidth widthResized))) -> do
    imageOrig <- CP.readImage source
    let imageOrig' = case imageOrig of
                       Left s -> error ("unexpected error: " ++ s)
                       Right image -> CP.convertRGB8 image
        sizeFactor :: Double
        sizeFactor = fromIntegral (CP.imageWidth imageOrig') / fromIntegral widthResized
        heightResized = round (fromIntegral (CP.imageHeight imageOrig') / sizeFactor)
        imageResized = CPS.resize CPS.defaultOptions widthResized heightResized imageOrig'
    CP.saveJpgImage 90 target $ CP.ImageRGB8 imageResized
    pure Empty

  (SaveTextFile, Tuple (Text t, String (StringWrapper s))) -> do
    T.writeFile s t
    pure Empty

  (CopyFile, Tuple (String (StringWrapper source), String (StringWrapper target))) -> do
    copyFile source target
    pure Empty

  (MakeDir, String (StringWrapper s)) -> do
    createDirectory s
    pure Empty

  (RunPandoc, Text contents) -> do
    html <- P.runIOorExplode (P.writeHtml5 P.def =<< P.readMarkdown (P.def { P.readerExtensions = P.extensionsFromList [ P.Ext_raw_html ] }) contents)
    pure $ Text $ TL.toStrict $ B.renderHtml html

  _ ->
    error ("unexpected combination of function and argument type; got function " ++ show f ++ " with argument " ++ show x)
